@@ -24,19 +24,20 @@ from collections import defaultdict
 from datetime import datetime
 from operator import attrgetter
 from time import time
-from typing import TYPE_CHECKING, List, Tuple
+from typing import TYPE_CHECKING, Any, Callable, List, Tuple
 from urllib.parse import quote
 
 # Using `from elasticsearch import *` would break elasticsearch mocking used in unit test.
 import elasticsearch
 import pendulum
-from elasticsearch_dsl import Search
+from elasticsearch.exceptions import ElasticsearchException, NotFoundError
 
 from airflow.configuration import conf
 from airflow.exceptions import AirflowProviderDeprecationWarning
 from airflow.models.dagrun import DagRun
 from airflow.models.taskinstance import TaskInstance
 from airflow.providers.elasticsearch.log.es_json_formatter import ElasticsearchJSONFormatter
+from airflow.providers.elasticsearch.log.es_response import ElasticSearchResponse, Hit
 from airflow.utils import timezone
 from airflow.utils.log.file_task_handler import FileTaskHandler
 from airflow.utils.log.logging_mixin import ExternalLoggingMixin, LoggingMixin
@@ -54,12 +55,12 @@ USE_PER_RUN_LOG_ID = hasattr(DagRun, "get_log_template")
 
 class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMixin):
     """
-    ElasticsearchTaskHandler is a python log handler that
-    reads logs from Elasticsearch. Note that Airflow does not handle the indexing
-    of logs into Elasticsearch. Instead, Airflow flushes logs
-    into local files. Additional software setup is required
-    to index the logs into Elasticsearch, such as using
-    Filebeat and Logstash.
+    ElasticsearchTaskHandler is a python log handler that reads logs from Elasticsearch.
+
+    Note that Airflow does not handle the indexing of logs into Elasticsearch. Instead,
+    Airflow flushes logs into local files. Additional software setup is required to index
+    the logs into Elasticsearch, such as using Filebeat and Logstash.
+
     To efficiently query and sort Elasticsearch results, this handler assumes each
     log message has a field `log_id` consists of ti primary keys:
     `log_id = {dag_id}-{task_id}-{execution_date}-{try_number}`
@@ -67,6 +68,10 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
     which is a unique integer indicates log message's order.
     Timestamps here are unreliable because multiple log messages
     might have the same timestamp.
+
+    :param base_log_folder: base folder to store logs locally
+    :param log_id_template: log id template
+    :param host: Elasticsearch host name
     """
 
     PAGE = 0
@@ -92,11 +97,6 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
         filename_template: str | None = None,
         log_id_template: str | None = None,
     ):
-        """
-        :param base_log_folder: base folder to store logs locally
-        :param log_id_template: log id template
-        :param host: Elasticsearch host name
-        """
         es_kwargs = es_kwargs or {}
         super().__init__(base_log_folder, filename_template)
         self.closed = False
@@ -123,6 +123,8 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
 
         self.formatter: logging.Formatter
         self.handler: logging.FileHandler | logging.StreamHandler  # type: ignore[assignment]
+        self._doc_type_map: dict[Any, Any] = {}
+        self._doc_type: list[Any] = []
 
     def _render_log_id(self, ti: TaskInstance, try_number: int) -> str:
         with create_session() as session:
@@ -170,8 +172,7 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
     @staticmethod
     def _clean_date(value: datetime | None) -> str:
         """
-        Clean up a date value so that it is safe to query in elasticsearch
-        by removing reserved characters.
+        Clean up a date value so that it is safe to query in elasticsearch by removing reserved characters.
 
         https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-query-string-query.html#_reserved_characters
         """
@@ -209,12 +210,9 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
 
         offset = metadata["offset"]
         log_id = self._render_log_id(ti, try_number)
-
         logs = self.es_read(log_id, offset, metadata)
         logs_by_host = self._group_logs_by_host(logs)
-
         next_offset = offset if not logs else attrgetter(self.offset_field)(logs[-1])
-
         # Ensure a string here. Large offset numbers will get JSON.parsed incorrectly
         # on the client. Sending as a string prevents this issue.
         # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Number/MAX_SAFE_INTEGER
@@ -259,7 +257,6 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
             return "\n".join(self._format_msg(lines[i]) for i in range(log_range))
 
         message = [(host, concat_logs(hosted_log)) for host, hosted_log in logs_by_host.items()]
-
         return message, metadata
 
     def _format_msg(self, log_line):
@@ -277,41 +274,44 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
         # Just a safe-guard to preserve backwards-compatibility
         return log_line.message
 
-    def es_read(self, log_id: str, offset: str, metadata: dict) -> list:
+    def es_read(self, log_id: str, offset: int | str, metadata: dict) -> list | ElasticSearchResponse:
         """
-        Returns the logs matching log_id in Elasticsearch and next offset.
-        Returns '' if no log is found or there was an error.
+        Return the logs matching log_id in Elasticsearch and next offset or ''.
 
         :param log_id: the log_id of the log to read.
         :param offset: the offset start to read log from.
         :param metadata: log metadata, used for steaming log download.
         """
-        # Offset is the unique key for sorting logs given log_id.
-        search = (
-            Search(index=self.index_patterns, using=self.client)
-            .query("match_phrase", log_id=log_id)
-            .sort(self.offset_field)
-        )
+        query: dict[Any, Any] = {
+            "query": {
+                "bool": {
+                    "filter": [{"range": {self.offset_field: {"gt": int(offset)}}}],
+                    "must": [{"match_phrase": {"log_id": log_id}}],
+                }
+            }
+        }
 
-        search = search.filter("range", **{self.offset_field: {"gt": int(offset)}})
-        max_log_line = search.count()
-        if "download_logs" in metadata and metadata["download_logs"] and "max_offset" not in metadata:
-            try:
-                if max_log_line > 0:
-                    metadata["max_offset"] = attrgetter(self.offset_field)(
-                        search[max_log_line - 1].execute()[-1]
-                    )
-                else:
-                    metadata["max_offset"] = 0
-            except Exception:
-                self.log.exception("Could not get current log size with log_id: %s", log_id)
+        try:
+            max_log_line = self.client.count(index=self.index_patterns, body=query)["count"]
+        except NotFoundError as e:
+            self.log.exception("The target index pattern %s does not exist", self.index_patterns)
+            raise e
+        except ElasticsearchException as e:
+            self.log.exception("Could not get current log size with log_id: %s", log_id)
+            raise e
 
-        logs = []
+        logs: list[Any] | ElasticSearchResponse = []
         if max_log_line != 0:
             try:
-
-                logs = search[self.MAX_LINE_PER_PAGE * self.PAGE : self.MAX_LINE_PER_PAGE].execute()
-            except Exception:
+                query.update({"sort": [self.offset_field]})
+                res = self.client.search(
+                    index=self.index_patterns,
+                    body=query,
+                    size=self.MAX_LINE_PER_PAGE,
+                    from_=self.MAX_LINE_PER_PAGE * self.PAGE,
+                )
+                logs = ElasticSearchResponse(self, res)
+            except elasticsearch.exceptions.ElasticsearchException:
                 self.log.exception("Could not read log with log_id: %s", log_id)
 
         return logs
@@ -415,6 +415,99 @@ class ElasticsearchTaskHandler(FileTaskHandler, ExternalLoggingMixin, LoggingMix
         """Whether we can support external links."""
         return bool(self.frontend)
 
+    def _resolve_nested(self, hit: dict[Any, Any], parent_class=None) -> type[Hit]:
+        """
+        Resolves nested hits from Elasticsearch by iteratively navigating the `_nested` field.
+        The result is used to fetch the appropriate document class to handle the hit.
+
+        This method can be used with nested Elasticsearch fields which are structured
+        as dictionaries with "field" and "_nested" keys.
+        """
+        doc_class = Hit
+
+        nested_path: list[str] = []
+        nesting = hit["_nested"]
+        while nesting and "field" in nesting:
+            nested_path.append(nesting["field"])
+            nesting = nesting.get("_nested")
+        nested_path_str = ".".join(nested_path)
+
+        if hasattr(parent_class, "_index"):
+            nested_field = parent_class._index.resolve_field(nested_path_str)
+
+        if nested_field is not None:
+            return nested_field._doc_class
+
+        return doc_class
+
+    def _get_result(self, hit: dict[Any, Any], parent_class=None) -> Hit:
+        """
+        This method processes a hit (i.e., a result) from an Elasticsearch response and transforms it into an
+        appropriate class instance.
+
+        The transformation depends on the contents of the hit. If the document in hit contains a nested field,
+        the '_resolve_nested' method is used to determine the appropriate class (based on the nested path).
+        If the hit has a document type that is present in the '_doc_type_map', the corresponding class is
+        used. If not, the method iterates over the '_doc_type' classes and uses the first one whose '_matches'
+        method returns True for the hit.
+
+        If the hit contains any 'inner_hits', these are also processed into 'ElasticSearchResponse' instances
+        using the determined class.
+
+        Finally, the transformed hit is returned. If the determined class has a 'from_es' method, this is
+        used to transform the hit
+
+        An example of the hit argument:
+
+        {'_id': 'jdeZT4kBjAZqZnexVUxk',
+         '_index': '.ds-filebeat-8.8.2-2023.07.09-000001',
+         '_score': 2.482621,
+         '_source': {'@timestamp': '2023-07-13T14:13:15.140Z',
+                     'asctime': '2023-07-09T07:47:43.907+0000',
+                     'container': {'id': 'airflow'},
+                     'dag_id': 'example_bash_operator',
+                     'ecs': {'version': '8.0.0'},
+                     'execution_date': '2023_07_09T07_47_32_000000',
+                     'filename': 'taskinstance.py',
+                     'input': {'type': 'log'},
+                     'levelname': 'INFO',
+                     'lineno': 1144,
+                     'log': {'file': {'path': "/opt/airflow/Documents/GitHub/airflow/logs/
+                     dag_id=example_bash_operator'/run_id=owen_run_run/
+                     task_id=run_after_loop/attempt=1.log"},
+                             'offset': 0},
+                     'log.offset': 1688888863907337472,
+                     'log_id': 'example_bash_operator-run_after_loop-owen_run_run--1-1',
+                     'message': 'Dependencies all met for dep_context=non-requeueable '
+                                'deps ti=<TaskInstance: '
+                                'example_bash_operator.run_after_loop owen_run_run '
+                                '[queued]>',
+                     'task_id': 'run_after_loop',
+                     'try_number': '1'},
+         '_type': '_doc'}
+        """
+        doc_class = Hit
+        dt = hit.get("_type")
+
+        if "_nested" in hit:
+            doc_class = self._resolve_nested(hit, parent_class)
+
+        elif dt in self._doc_type_map:
+            doc_class = self._doc_type_map[dt]
+
+        else:
+            for doc_type in self._doc_type:
+                if hasattr(doc_type, "_matches") and doc_type._matches(hit):
+                    doc_class = doc_type
+                    break
+
+        for t in hit.get("inner_hits", ()):
+            hit["inner_hits"][t] = ElasticSearchResponse(self, hit["inner_hits"][t], doc_class=doc_class)
+
+        # callback should get the Hit class if "from_es" is not defined
+        callback: type[Hit] | Callable[..., Any] = getattr(doc_class, "from_es", doc_class)
+        return callback(hit)
+
 
 def getattr_nested(obj, item, default):
     """
